"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

# -*- coding: utf-8 -*-
from libs import base
from constants import *
from alerts import namespace_stats
from helpers.info_dict import InfoDict
from helpers.debugger import Debugger
from utilities import connection_manager
from distutils import version

from flaskapp import app
logger = app.logger

debugger = Debugger()

class Namespace(object):
    """Namespace contains all information about namespace"""

    def __new__(cls, name, ip, port, cluster_id, build):

        #if base.is_node_available(ip, port) is False:
        #    return None
        return super(Namespace, cls).__new__(cls, name, ip, port, cluster_id, build)

    def __init__(self, name, ip, port, cluster_id, build):

        self._name = name
        self._ip = ip
        self._port = port
        self._address = "%s:%s" % (ip, port)
        self._cluster_id = cluster_id
        self._build = build

        self._statistics = {}
        self._raw_statistics = {}
        self._stats_alert = namespace_stats.NamespaceStats(
            self._cluster_id, self._address, self._name)
        self._modified_statistics = {}
        self._modified_raw_statistics = {}
        self._index_statistics = {}
        self._modified_index_statistics = {}
        self.configuration_params = ["type", "storage-engine",
                                        "file", "storage-engine.file",
                                        "filesize", "storage-engine.filesize",
                                        "load-at-startup",
                                        "data-in-memory",
                                        "defrag-period",
                                        "defrag-lwm-pct", "storage-engine.defrag-lwm-pct",
                                        "defrag-max-blocks",
                                        "defrag-startup-minimum", "storage-engine.defrag-startup-minimum",
                                        "write-block-size", "storage-engine.write-block-size",
                                        "ticker-interval",
                                        "replication-factor",
                                        "low-water-pct",
                                        "high-water-memory-pct",
                                        "high-water-disk-pct",
                                        "high-water-pct",
                                        "evict-pct",
                                        "stop-writes-pct",
                                        "memory-size",
                                        "default-ttl",
                                        "max-ttl",
                                        "allow-versions",
                                        "single-bin",
                                        "node_status",
                                        "migrate-sleep", "migrate-order",
                                    ]

    def get_configurations(self):

        return self.get_subset_of_statistics(self.configuration_params)

    def get_configurations_without_ns(self):

        configs_without_ns = InfoDict()

        stats = self.get_statistics()
        for attr in self.configuration_params:
            value = stats.get(attr, None)
            if value is not None:
                configs_without_ns[attr] = value
        return configs_without_ns

    def get_subset_of_configurations(self, attrs=[]):

        subset = InfoDict()
        configurations = self.get_configurations()

        for attr in attrs:
            subset[attr] = configurations.get(attr, NOT_SUPPORTED)
        return subset

    def update_statistics(self, user=None, password=None, node=None):

        if node is not None:
            sock = node.get_socket_from_queue()

            # statistics = base.get_namespace_statistics(self._name, self._ip, self._port, clear_cache=True, user=user, password=password, sock=sock)
            statistics = base.get_namespace_statistics(self._name, self._ip, self._port, user=user, password=password, sock=sock)
            if statistics.get('node_status',None) is 'off':
                new_sock = connection_manager.create_socket(self._ip, self._port, user=user, password=password)
                if new_sock is not None:
                    sock = new_sock
                    # statistics = base.get_namespace_statistics(self._name, self._ip, self._port, clear_cache=True, user=user, password=password, sock=sock)
                    statistics = base.get_namespace_statistics(self._name, self._ip, self._port, user=user, password=password, sock=sock)
            node.put_socket_in_queue(sock)
        else:
            # statistics = base.get_namespace_statistics(self._name, self._ip, self._port, clear_cache=True, user=user, password=password)
            statistics = base.get_namespace_statistics(self._name, self._ip, self._port, user=user, password=password)

        raw_statistics = statistics.copy()

        # for version 3.9 and after
        if version.LooseVersion(self._build) >= version.LooseVersion("3.8.4-110"):
            statistics = self.assign_stat_aliases(statistics)

        self._modified_statistics = dict(set(statistics.items()) - set(self._statistics.items()))
        self._modified_raw_statistics = dict(set(raw_statistics.items()) - set(self._raw_statistics.items()))
        self._modified_statistics['node_status'] = self._statistics.get('node_status', 'off')
        self._modified_raw_statistics['node_status'] = self._raw_statistics.get('node_status', 'off')
        self._statistics = statistics
        self._raw_statistics = raw_statistics

    def set_configuration(self, parameter, value, user=None, password=None, node=None):

        if node is not None:
            sock = node.get_socket_from_queue()
            status = base.set_namespace_configuration(self._name, self._ip, self._port, str(parameter), str(value), user=user, password=password, sock=sock)
            if status is False:
                new_sock = connection_manager.create_socket(self._ip, self._port,user=user, password=password)
                if new_sock is not None:
                    sock = new_sock
                    status = base.set_namespace_configuration(self._name, self._ip, self._port, str(parameter), str(value), user=user, password=password, sock=sock)
            node.put_socket_in_queue(sock)
        else:
            status = base.set_namespace_configuration(self._name, self._ip, self._port,
                                        str(parameter), str(value), user=user, password=password)
        return status

    def set_configurations(self, edit_parameters, user=None, password=None, node=None):

        error_list = list()
        for key, value in list(edit_parameters.items()):
            if self.set_configuration(key, value, user=user, password=password, node=node) is False:
                error_list.append(key)
        self.update_statistics(user=user, password=password, node=node)

        return error_list

    def get_statistics(self, raw=False):

        if raw:
            return self._raw_statistics
        else:
            return self._statistics

    def get_modified_statistics(self, raw=False, user=None, password=None, node=None):

        self.update_statistics(user=user, password=password, node=node)

        if raw:
            return self._modified_raw_statistics
        else:
            return self._modified_statistics

    def assign_stat_aliases(self, stats):
        # for version 3.9 and after
        stats["type"] = stats.get("storage-engine", "memory")
        stats["file"] = stats.get("storage-engine.file", None)
        stats["filesize"] = stats.get("storage-engine.filesize", None)
        stats["defrag-lwm-pct"] = stats.get("storage-engine.defrag-lwm-pct", None)
        stats["defrag-startup-minimum"] = stats.get("storage-engine.defrag-startup-minimum", None)
        stats["write-block-size"] = stats.get("storage-engine.write-block-size", None)

        if stats.get("device_total_bytes", None) is not None:
            stats["used-bytes-disk"] = stats.get("device_used_bytes", None)
            stats["total-bytes-disk"] = stats.get("device_total_bytes", None)
            stats["free-bytes-disk"] = str(int(stats["total-bytes-disk"]) - int(stats["used-bytes-disk"]))
            stats["free-pct-disk"] = stats.get("device_free_pct", None)
            stats["available_pct"] = stats.get("device_available_pct", None)
            stats["free_pct"] = stats.get("device_free_pct", None)

        stats["total-bytes-memory"] = stats["memory-size"]
        stats["data-in-memory"] = (stats.get("storage-engine", "memory") == "memory")

        stats["available-bin-names"] = stats["available_bin_names"]
        stats["current-time"] = stats["current_time"]
        stats["data-in-memory"] = stats["data-in-memory"]
        stats["evicted-objects"] = stats["evicted_objects"]
        stats["expired-objects"] = stats["expired_objects"]
        stats["hwm-breached"] = stats["hwm_breached"]
        stats["master-objects"] = stats["master_objects"]
        stats["master-sub-objects"] = stats["master_sub_objects"]
        stats["max-void-time"] = stats["max_void_time"]
        stats["free-pct-memory"] = stats["memory_free_pct"]
        stats["used-bytes-memory"] = stats["memory_used_bytes"]
        stats["data-used-bytes-memory"] = stats["memory_used_data_bytes"]
        stats["index-used-bytes-memory"] = stats["memory_used_index_bytes"]
        stats["sindex-used-bytes-memory"] = stats["memory_used_sindex_bytes"]
        stats["migrate-record-receives"] = stats["migrate_record_receives"]
        stats["migrate-record-retransmits"] = stats["migrate_record_retransmits"]
        stats["migrate-records-skipped"] = stats["migrate_records_skipped"]
        stats["migrate-records-transmitted"] = stats["migrate_records_transmitted"]
        stats["migrate-rx-instance-count"] = stats["migrate_rx_instances"]
        stats["migrate-rx-partitions-active"] = stats["migrate_rx_partitions_active"]
        stats["migrate-rx-partitions-initial"] = stats["migrate_rx_partitions_initial"]
        stats["migrate-rx-partitions-remaining"] = stats["migrate_rx_partitions_remaining"]
        stats["migrate-tx-instance-count"] = stats["migrate_tx_instances"]
        stats["migrate-tx-partitions-active"] = stats["migrate_tx_partitions_active"]
        stats["migrate-tx-partitions-imbalance"] = stats["migrate_tx_partitions_imbalance"]
        stats["migrate-tx-partitions-initial"] = stats["migrate_tx_partitions_initial"]
        stats["migrate-tx-partitions-remaining"] = stats["migrate_tx_partitions_remaining"]
        stats["non-expirable-objects"] = stats["non_expirable_objects"]
        stats["nsup-cycle-duration"] = stats["nsup_cycle_duration"]
        stats["nsup-cycle-sleep-pct"] = stats["nsup_cycle_sleep_pct"]
        stats["prole-objects"] = stats["prole_objects"]
        stats["prole-sub-objects"] = stats["prole_sub_objects"]
        stats["set-deleted-objects"] = stats["set_deleted_objects"]
        stats["stop-writes"] = stats["stop_writes"]
        stats["defrag-lwm-pct"] = stats.get("storage-engine.defrag-lwm-pct", None)
        stats["defrag-queue-min"] = stats.get("storage-engine.defrag-queue-min", None)
        stats["defrag-sleep"] = stats.get("storage-engine.defrag-sleep", None)
        stats["defrag-startup-minimum"] = stats.get("storage-engine.defrag-startup-minimum", None)
        stats["filesize"] = stats.get("storage-engine.filesize", None)
        stats["flush-max-ms"] = stats.get("storage-engine.flush-max-ms", None)
        stats["fsync-max-sec"] = stats.get("storage-engine.fsync-max-sec", None)
        stats["max-write-cache"] = stats.get("storage-engine.max-write-cache", None)
        stats["min-avail-pct"] = stats.get("storage-engine.min-avail-pct", None)
        stats["post-write-queue"] = stats.get("storage-engine.post-write-queue", None)
        stats["write-block-size"] = stats.get("storage-engine.write-block-size", None)
        stats["sub-objects"] = stats["sub_objects"]

        return stats

    def get_updated_statistics(self, raw=False, user=None, password=None, node=None):

        if node is not None:
            sock = node.get_socket_from_queue()
            statistics = base.get_namespace_statistics(self._name, self._ip, self._port, user=user, password=password, sock=sock)
            if statistics.get('node_status',None) is 'off':
                new_sock = connection_manager.create_socket(self._ip, self._port, user=user, password=password)
                if new_sock is not None:
                    sock = new_sock
                    statistics = base.get_namespace_statistics(self._name, self._ip, self._port, user=user, password=password, sock=sock)
            node.put_socket_in_queue(sock)
        else:
            statistics = base.get_namespace_statistics(self._name,
                self._ip, self._port, user=user, password=password)

        raw_statistics = statistics.copy()

        # for version 3.9 and after
        if version.LooseVersion(self._build) >= version.LooseVersion("3.8.4-110"):
            statistics = self.assign_stat_aliases(statistics)
        else:
            statistics["xdr_write_success"] = statistics.get("stat_recs_shipped_ok", "0")

        self._modified_statistics = set(statistics.items()) - set(self._statistics.items())
        self._modified_raw_statistics = set(raw_statistics.items()) - set(self._raw_statistics.items())
        self._statistics = statistics
        self._raw_statistics = raw_statistics

        ##    * Unit test cases *
        ##from random import randrange

        ##free_pct_disk = ["2", "5", "10"]
        ##self._statistics['free-pct-disk'] = free_pct_disk[randrange(0, 3)]

        ##available_pct = ["5", "10", "20"]
        ##self._statistics["available_pct"] = available_pct[randrange(0, 3)]

        ##disks = ["80", "90", "100"]
        ##self._statistics["used-disk-pct"] = disks[randrange(0, 3)]
        ##self._statistics["high-water-disk-pct"] = "80"
        ##self._statistics["stop-writes-pct"] = "90"

        ##memory = ["80", "90", "100"]
        ##self._statistics["used-memory-pct"] = memory[randrange(0, 3)]
        ##self._statistics["high-water-memory-pct"] = "80"
        ##self._statistics["stop-writes-pct"] = "90"

        if raw:
            return self._raw_statistics
        else:
            return self._statistics

    def get_subset_of_statistics(self, attrs=[]):

        subset = InfoDict()

        stats = self.get_statistics()
        for attr in attrs:
            subset[attr] = stats.get(attr, NOT_SUPPORTED)
        return subset

    def get_statistic(self, attr):

        stats = self.get_statistics()
        return stats.get(attr, NOT_SUPPORTED)

    def get_disk(self):

        return self.get_statistic("used-bytes-disk"), self.get_statistic("total-bytes-disk")

    def get_disk_stats(self):

        attrs = ["used-bytes-disk", "total-bytes-disk"]
        stats = self.get_subset_of_statistics(attrs)
        try:
            stats["used-bytes-disk"] = int(stats["used-bytes-disk"])
            stats["total-bytes-disk"] = int(stats["total-bytes-disk"])
            stats["free-bytes-disk"] = max(0, stats["total-bytes-disk"] - stats["used-bytes-disk"])
        except:
            stats["free-bytes-disk"] = NOT_SUPPORTED
        return stats

    def get_memory_stats(self):

        attrs = ["used-bytes-memory", "total-bytes-memory"]
        stats = self.get_subset_of_statistics(attrs)
        try:
            stats["used-bytes-memory"] = int(stats["used-bytes-memory"])
            stats["total-bytes-memory"] = int(stats["total-bytes-memory"])
            stats["free-bytes-memory"] = max(0, stats["total-bytes-memory"] - stats["used-bytes-memory"])
        except:
            stats["free-bytes-memory"] = NOT_SUPPORTED
        return stats

    def get_memory(self):

        return self.get_statistic("used-bytes-memory"), self.get_statistic("total-bytes-memory")

    def get_sets_info(self, user=None, password=None):

        """
        This methos is no more in USE. Can be removed from source
        at appropriate time.
        """
        self._sets_info = base.get_namespace_sets_info(self._name, user=user, password=password)
        return self._sets_info

    def update_alerts(self, user=None, password=None, node=None):

        self._stats_alert.generate_alerts(self.get_updated_statistics(raw=False, user=user, password=password, node=node))

    def get_index_statistics(self, index_name, stats_type="all", user=None, password=None, node=None):

        if node is not None:
            sock = node.get_socket_from_queue()
            index_statistics = base.get_index_stats(self._ip, self._port, self._name, index_name, user=user, password=password, sock=sock)
            if index_statistics.get('node_status',None) is 'off':
                new_sock = connection_manager.create_socket(self._ip, self._port, user=user, password=password)
                if new_sock is not None:
                    sock= new_sock
                    index_statistics = base.get_index_stats(self._ip, self._port, self._name, index_name, user=user, password=password, sock=sock)
            node.put_socket_in_queue(sock)
        else:
            index_statistics = base.get_index_stats(self._ip, self._port, self._name, index_name, user=user, password=password)
        self._modified_index_statistics = dict(set(index_statistics.items()) - set(self._index_statistics.items()))
        self._index_statistics = index_statistics

        self._modified_index_statistics['node_status'] = self._index_statistics.get('node_status', 'off')

        return self._index_statistics if stats_type == "all" else self._modified_index_statistics

    def get_debug_status(self):
        return debugger.status()
